/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "modelconnector.h"
#include "model.h"
#include "jsonescaper.h"
#include "postgresqlescaper.h"


namespace morm
{

ModelConnector::ModelConnector()
{
	expression = nullptr;
	out_stream = nullptr;
	json_escaper = nullptr;
	postgresql_escaper = nullptr;

	expression_allocated = false;
	out_stream_allocated = false;
	json_escaper_allocated = false;
	postgresql_escaper_allocated = false;

	assigned_db_type = MORM_DB_TYPE_POSTGRESQL;
}

ModelConnector::~ModelConnector()
{
	deallocate_expression();
	deallocate_stream();
	deallocate_json_escaper();
	deallocate_postgresql_escaper();
}

void ModelConnector::deallocate_expression()
{
	if( expression_allocated )
	{
		delete expression;
		expression = nullptr;
		expression_allocated = false;
	}
}

void ModelConnector::allocate_default_expression()
{
	deallocate_expression();
	expression = new Expression();
	expression_allocated = true;
}


void ModelConnector::deallocate_stream()
{
	if( out_stream_allocated )
	{
		delete out_stream;
		out_stream = nullptr;
		out_stream_allocated = false;
	}
}


void ModelConnector::allocate_default_stream()
{
	deallocate_stream();
	out_stream = new PT::TextStream();
	out_stream_allocated = true;
}


void ModelConnector::deallocate_json_escaper()
{
	if( json_escaper_allocated )
	{
		delete json_escaper;
		json_escaper = nullptr;
		json_escaper_allocated = false;
	}
}


void ModelConnector::allocate_default_json_escaper()
{
	deallocate_json_escaper();
	json_escaper = new JSONEscaper();
	json_escaper_allocated = true;
}


void ModelConnector::deallocate_postgresql_escaper()
{
	if( postgresql_escaper_allocated )
	{
		delete postgresql_escaper;
		postgresql_escaper = nullptr;
		postgresql_escaper_allocated = false;
	}
}


void ModelConnector::allocate_default_postgresql_escaper()
{
	deallocate_postgresql_escaper();
	postgresql_escaper = new PostgreSQLEscaper();
	postgresql_escaper_allocated = true;
}


void ModelConnector::allocate_default_stream_if_needed()
{
	if( !out_stream )
		allocate_default_stream();
}

void ModelConnector::allocate_default_expression_if_needed()
{
	if( !expression )
		allocate_default_expression();
}

void ModelConnector::allocate_default_json_escaper_if_needed()
{
	if( !json_escaper )
		allocate_default_json_escaper();
}

void ModelConnector::allocate_default_postgresql_escaper_if_needed()
{
	if( !postgresql_escaper )
		allocate_default_postgresql_escaper();
}

void ModelConnector::allocate_default_db_escaper_if_needed()
{
	if( assigned_db_type == MORM_DB_TYPE_POSTGRESQL )
	{
		allocate_default_postgresql_escaper_if_needed();
	}
}



void ModelConnector::set_expression(Expression & expression)
{
	deallocate_expression();
	this->expression = &expression;
}

void ModelConnector::set_stream(PT::TextStream & stream)
{
	deallocate_stream();
	this->out_stream = &stream;
}

Expression * ModelConnector::get_expression()
{
	return expression;
}

PT::TextStream * ModelConnector::get_stream()
{
	return out_stream;
}



void ModelConnector::to_json(PT::TextStream & stream, Model & model)
{
	allocate_default_expression_if_needed();
	allocate_default_json_escaper_if_needed();

	expression->set_json_escaper(*json_escaper);
	expression->to_json(stream, model);
}


void ModelConnector::insert(PT::TextStream & stream, Model & model, int db_type)
{
	allocate_default_expression_if_needed();
	allocate_default_db_escaper_if_needed(); // FIX ME what about db_type parameter

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->insert(stream, model, db_type);
}


void ModelConnector::update(PT::TextStream & stream, Model & model, int db_type)
{
	allocate_default_expression_if_needed();
	allocate_default_db_escaper_if_needed(); // FIX ME what about db_type parameter

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->update(stream, model, db_type);
}


void ModelConnector::insert(PT::TextStream & stream, Model & model)
{
	allocate_default_expression_if_needed();
	allocate_default_db_escaper_if_needed();

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->insert(stream, model, assigned_db_type);
}


void ModelConnector::update(PT::TextStream & stream, Model & model)
{
	allocate_default_expression_if_needed();
	allocate_default_db_escaper_if_needed();

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->update(stream, model, assigned_db_type);
}



void ModelConnector::to_json(Model & model)
{
	allocate_default_expression_if_needed();
	allocate_default_stream_if_needed();
	allocate_default_json_escaper_if_needed();

	expression->set_json_escaper(*json_escaper);
	expression->to_json(*out_stream, model);
}


void ModelConnector::insert(Model & model, int db_type)
{
	allocate_default_expression_if_needed();
	allocate_default_stream_if_needed();
	allocate_default_db_escaper_if_needed(); // FIX ME what about db_type parameter?

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->insert(*out_stream, model, db_type);
}


void ModelConnector::update(Model & model, int db_type)
{
	allocate_default_expression_if_needed();
	allocate_default_stream_if_needed();
	allocate_default_db_escaper_if_needed(); // FIX ME what about db_type parameter?

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->update(*out_stream, model, db_type);
}



void ModelConnector::insert(Model & model)
{
	allocate_default_expression_if_needed();
	allocate_default_stream_if_needed();
	allocate_default_db_escaper_if_needed();

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->insert(*out_stream, model, assigned_db_type);
}


void ModelConnector::update(Model & model)
{
	allocate_default_expression_if_needed();
	allocate_default_stream_if_needed();
	allocate_default_db_escaper_if_needed();

	expression->set_db_escaper(*postgresql_escaper); //FIX ME select correct db escaper
	expression->update(*out_stream, model, assigned_db_type);
}






}
